@@ -1,7 +1,7 @@
 import os
 import unittest
 from tpv.rules import gateway
-from . import mock_galaxy
+from tpv.core.test import mock_galaxy
 from galaxy.jobs.mapper import JobMappingException
 
 
@@ -9,7 +9,7 @@ class TestMapperBasic(unittest.TestCase):
 
     @staticmethod
     def _map_to_destination(tool):
-        galaxy_app = mock_galaxy.App()
+        galaxy_app = mock_galaxy.App(job_conf=os.path.join(os.path.dirname(__file__), 'fixtures/job_conf.yml'))
         job = mock_galaxy.Job()
         user = mock_galaxy.User('gargravarr', 'fairycake@vortex.org')
         tpv_config = os.path.join(os.path.dirname(__file__), 'fixtures/mapping-basic.yml')
@@ -45,3 +45,13 @@ class TestMapperBasic(unittest.TestCase):
         tool = mock_galaxy.Tool('regex_t_test')
         destination = self._map_to_destination(tool)
         self.assertEqual(destination.id, "local")
+
+    def test_map_abstract_tool_should_fail(self):
+        tool = mock_galaxy.Tool('my_abstract_tool')
+        with self.assertRaisesRegex(JobMappingException, "This entity is abstract and cannot be mapped"):
+            self._map_to_destination(tool)
+
+    def test_map_concrete_descendant_should_succeed(self):
+        tool = mock_galaxy.Tool('my_concrete_tool')
+        destination = self._map_to_destination(tool)
+        self.assertEqual(destination.id, "local")
